from importlib import import_module

from bs4 import BeautifulSoup

from .. import hooks
from ..conf import settings


class MarkupPipeline:
    """small framework for extending parser"""

    def extend_markdown(self, md):
        for extension in settings.MISAGO_MARKUP_EXTENSIONS:
            module = import_module(extension)
            if hasattr(module, "extend_markdown"):
                hook = getattr(module, "extend_markdown")
                hook.extend_markdown(md)

        for extension in hooks.markdown_extensions:
            extension(md)

        return md

    def process_result(self, result):
        soup = BeautifulSoup(result["parsed_text"], "html5lib")
        for extension in settings.MISAGO_MARKUP_EXTENSIONS:
            module = import_module(extension)
            if hasattr(module, "clean_parsed"):
                hook = getattr(module, "clean_parsed")
                hook.process_result(result, soup)

        for extension in hooks.parsing_result_processors:
            extension(result, soup)

        souped_text = str(soup.body).strip()[6:-7]
        result["parsed_text"] = souped_text.strip()
        return result


pipeline = MarkupPipeline()



from ..parser import parse


def test_strikethrough_markdown(request_mock, user, snapshot):
    text = "Lorem ~~ipsum~~ dolor met!"
    result = parse(text, request_mock, user, minify=False)
    snapshot.assert_match(result["parsed_text"])



from ..parser import parse


def test_html_is_escaped(request_mock, user, snapshot):
    text = "Lorem <strong>ipsum!</strong>"
    result = parse(text, request_mock, user, minify=True)
    snapshot.assert_match(result["parsed_text"])


def test_parsed_text_is_minified(request_mock, user, snapshot):
    text = """
Lorem **ipsum** dolor met.

Sit amet elit.
"""
    result = parse(text, request_mock, user, minify=True)
    snapshot.assert_match(result["parsed_text"])



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_code_in_quote_bbcode_is_escaped 1"
] = """<aside class="quote-block">
<div class="quote-heading"></div>
<blockquote class="quote-body">
<p>&lt;script&gt;alert("!")&lt;/script&gt;</p>
</blockquote>
</aside>"""

snapshots[
    "test_code_in_quote_markdown_is_escaped 1"
] = """<blockquote>
<p>&lt;script&gt;alert("!")&lt;/script&gt;</p>
</blockquote>"""

snapshots[
    "test_inline_code_is_escaped 1"
] = '<p><code>&lt;script&gt;alert("!")&lt;/script&gt;</code></p>'

snapshots["test_text_is_escaped 1"] = '<p>&lt;script&gt;alert("!")&lt;/script&gt;</p>'



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_html_is_escaped 1"] = "<p>Lorem &lt;strong&gt;ipsum!&lt;/strong&gt;</p>"

snapshots[
    "test_parsed_text_is_minified 1"
] = "<p>Lorem <strong>ipsum</strong> dolor met.</p><p>Sit amet elit.</p>"



import bleach
import markdown
from bs4 import BeautifulSoup
from django.http import Http404
from django.urls import resolve
from htmlmin.minify import html_minify
from markdown.extensions.fenced_code import FencedCodeExtension

from ..conf import settings
from .bbcode.code import CodeBlockExtension
from .bbcode.hr import BBCodeHRProcessor
from .bbcode.inline import bold, image, italics, underline, url
from .bbcode.quote import QuoteExtension
from .bbcode.spoiler import SpoilerExtension
from .md.shortimgs import ShortImagesExtension
from .md.strikethrough import StrikethroughExtension
from .mentions import add_mentions
from .pipeline import pipeline

MISAGO_ATTACHMENT_VIEWS = ("misago:attachment", "misago:attachment-thumbnail")


def parse(
    text,
    request,
    poster,
    allow_mentions=True,
    allow_links=True,
    allow_images=True,
    allow_blocks=True,
    force_shva=False,
    minify=True,
):
    """
    Message parser

    Utility for flavours to call

    Breaks text into paragraphs, supports code, spoiler and quote blocks,
    headers, lists, images, spoilers, text styles

    Returns dict object
    """
    md = md_factory(
        allow_links=allow_links, allow_images=allow_images, allow_blocks=allow_blocks
    )

    parsing_result = {
        "original_text": text,
        "parsed_text": "",
        "markdown": md,
        "mentions": [],
        "images": [],
        "internal_links": [],
        "outgoing_links": [],
    }

    # Parse text
    parsed_text = md.convert(text)

    # Clean and store parsed text
    parsing_result["parsed_text"] = parsed_text.strip()

    if allow_links:
        linkify_paragraphs(parsing_result)

    parsing_result = pipeline.process_result(parsing_result)

    if allow_mentions:
        add_mentions(request, parsing_result)

    if allow_links or allow_images:
        clean_links(request, parsing_result, force_shva)

    if minify:
        minify_result(parsing_result)
    return parsing_result


def md_factory(allow_links=True, allow_images=True, allow_blocks=True):
    """creates and configures markdown object"""
    md = markdown.Markdown(extensions=["markdown.extensions.nl2br"])

    # Remove HTML allowances
    md.preprocessors.deregister("html_block")
    md.inlinePatterns.deregister("html")

    # Remove references
    md.parser.blockprocessors.deregister("reference")
    md.inlinePatterns.deregister("reference")
    md.inlinePatterns.deregister("image_reference")
    md.inlinePatterns.deregister("short_reference")

    # Add [b], [i], [u]
    md.inlinePatterns.register(bold, "bb_b", 55)
    md.inlinePatterns.register(italics, "bb_i", 55)
    md.inlinePatterns.register(underline, "bb_u", 55)

    # Add ~~deleted~~
    strikethrough_md = StrikethroughExtension()
    strikethrough_md.extendMarkdown(md)

    if allow_links:
        # Add [url]
        md.inlinePatterns.register(url(md), "bb_url", 155)
    else:
        # Remove links
        md.inlinePatterns.deregister("link")
        md.inlinePatterns.deregister("autolink")
        md.inlinePatterns.deregister("automail")

    if allow_images:
        # Add [img]
        md.inlinePatterns.register(image(md), "bb_img", 145)
        short_images_md = ShortImagesExtension()
        short_images_md.extendMarkdown(md)
    else:
        # Remove images
        md.inlinePatterns.deregister("image_link")

    if allow_blocks:
        # Add [hr] and [quote] blocks
        md.parser.blockprocessors.register(BBCodeHRProcessor(md.parser), "bb_hr", 45)

        fenced_code = FencedCodeExtension(lang_prefix="")
        fenced_code.extendMarkdown(md)

        code_bbcode = CodeBlockExtension()
        code_bbcode.extendMarkdown(md)

        quote_bbcode = QuoteExtension()
        quote_bbcode.extendMarkdown(md)

        spoiler_bbcode = SpoilerExtension()
        spoiler_bbcode.extendMarkdown(md)
    else:
        # Remove blocks
        md.parser.blockprocessors.deregister("hashheader")
        md.parser.blockprocessors.deregister("setextheader")
        md.parser.blockprocessors.deregister("code")
        md.parser.blockprocessors.deregister("quote")
        md.parser.blockprocessors.deregister("hr")
        md.parser.blockprocessors.deregister("olist")
        md.parser.blockprocessors.deregister("ulist")

    return pipeline.extend_markdown(md)


def linkify_paragraphs(result):
    result["parsed_text"] = bleach.linkify(
        result["parsed_text"],
        callbacks=settings.MISAGO_BLEACH_CALLBACKS,
        skip_tags=["a", "code", "pre"],
        parse_email=True,
    )


def clean_links(request, result, force_shva=False):
    host = request.get_host()

    soup = BeautifulSoup(result["parsed_text"], "html5lib")
    for link in soup.find_all("a"):
        if is_internal_link(link["href"], host):
            link["href"] = clean_internal_link(link["href"], host)
            result["internal_links"].append(link["href"])
            link["href"] = clean_attachment_link(link["href"], force_shva)
        else:
            result["outgoing_links"].append(clean_link_prefix(link["href"]))
            link["href"] = assert_link_prefix(link["href"])
            link["rel"] = "external nofollow noopener"

        link["target"] = "_blank"

        if link.string:
            link.string = clean_link_prefix(link.string)

    for img in soup.find_all("img"):
        img["alt"] = clean_link_prefix(img["alt"])
        if is_internal_link(img["src"], host):
            img["src"] = clean_internal_link(img["src"], host)
            result["images"].append(img["src"])
            img["src"] = clean_attachment_link(img["src"], force_shva)
        else:
            result["images"].append(clean_link_prefix(img["src"]))
            img["src"] = assert_link_prefix(img["src"])

    # [6:-7] trims <body></body> wrap
    result["parsed_text"] = str(soup.body)[6:-7]


def is_internal_link(link, host):
    if link.startswith("/") and not link.startswith("//"):
        return True

    link = clean_link_prefix(link).lstrip("www.").lower()
    return link.lower().startswith(host.lstrip("www."))


def clean_link_prefix(link):
    if link.lower().startswith("https:"):
        link = link[6:]
    if link.lower().startswith("http:"):
        link = link[5:]
    if link.startswith("//"):
        link = link[2:]
    return link


def assert_link_prefix(link):
    if link.lower().startswith("https:"):
        return link
    if link.lower().startswith("http:"):
        return link
    if link.startswith("//"):
        return "http:%s" % link

    return "http://%s" % link


def clean_internal_link(link, host):
    link = clean_link_prefix(link)

    if link.lower().startswith("www."):
        link = link[4:]
    if host.lower().startswith("www."):
        host = host[4:]

    if link.lower().startswith(host):
        link = link[len(host) :]

    return link or "/"


def clean_attachment_link(link, force_shva=False):
    try:
        resolution = resolve(link)
        if not resolution.namespaces:
            return link
        url_name = ":".join(resolution.namespaces + [resolution.url_name])
    except (Http404, ValueError):
        return link

    if url_name in MISAGO_ATTACHMENT_VIEWS:
        if force_shva:
            link = "%s?shva=1" % link
        elif link.endswith("?shva=1"):
            link = link[:-7]
    return link


def minify_result(result):
    result["parsed_text"] = html_minify(result["parsed_text"])
    result["parsed_text"] = strip_html_head_body(result["parsed_text"])


def strip_html_head_body(parsed_text):
    # [25:-14] trims <html><head></head><body> and </body></html>
    return parsed_text[25:-14]



import re

from bs4 import BeautifulSoup
from django.contrib.auth import get_user_model

SUPPORTED_TAGS = ("h1", "h2", "h3", "h4", "h5", "h6", "div", "p")
USERNAME_RE = re.compile(r"@[0-9a-z]+", re.IGNORECASE)
MENTIONS_LIMIT = 24


def add_mentions(request, result):
    if "@" not in result["parsed_text"]:
        return

    mentions_dict = {}

    soup = BeautifulSoup(result["parsed_text"], "html5lib")

    elements = []
    for tagname in SUPPORTED_TAGS:
        if tagname in result["parsed_text"]:
            elements += soup.find_all(tagname)
    for element in elements:
        add_mentions_to_element(request, element, mentions_dict)

    result["parsed_text"] = str(soup.body)[6:-7].strip()
    result["mentions"] = list(filter(bool, mentions_dict.values()))


def add_mentions_to_element(request, element, mentions_dict):
    for item in element.contents:
        if item.name:
            if item.name != "a":
                add_mentions_to_element(request, item, mentions_dict)
        elif "@" in item.string:
            parse_string(request, item, mentions_dict)


def parse_string(request, element, mentions_dict):
    User = get_user_model()

    def replace_mentions(matchobj):
        if len(mentions_dict) >= MENTIONS_LIMIT:
            return matchobj.group(0)

        username = matchobj.group(0)[1:].strip().lower()

        if username not in mentions_dict:
            if username == request.user.slug:
                mentions_dict[username] = request.user
            else:
                try:
                    mentions_dict[username] = User.objects.get(slug=username)
                except User.DoesNotExist:
                    mentions_dict[username] = None

        if mentions_dict[username]:
            user = mentions_dict[username]
            return '<a href="%s">@%s</a>' % (user.get_absolute_url(), user.username)

        # we've failed to resolve user for username
        return matchobj.group(0)

    replaced_string = USERNAME_RE.sub(replace_mentions, element.string)
    element.replace_with(BeautifulSoup(replaced_string, "html.parser"))



from ..mentions import add_mentions


def test_util_replaces_mention_with_link_to_user_profile_in_parsed_text(
    request_mock, user
):
    parsing_result = {"parsed_text": f"<p>Hello, @{user.username}!</p>", "mentions": []}
    add_mentions(request_mock, parsing_result)
    assert parsing_result["parsed_text"] == (
        f'<p>Hello, <a href="{user.get_absolute_url()}">@{user.username}</a>!</p>'
    )


def test_util_adds_mention_to_parsig_result(request_mock, user):
    parsing_result = {"parsed_text": f"<p>Hello, @{user.username}!</p>", "mentions": []}
    add_mentions(request_mock, parsing_result)
    assert parsing_result["mentions"] == [user]


def test_mentions_arent_added_for_nonexisting_user(request_mock, user):
    parsing_result = {"parsed_text": f"<p>Hello, @OtherUser!</p>", "mentions": []}
    add_mentions(request_mock, parsing_result)
    assert parsing_result["parsed_text"] == "<p>Hello, @OtherUser!</p>"


def test_util_replaces_multiple_mentions_with_link_to_user_profiles_in_parsed_text(
    request_mock, user, other_user
):
    parsing_result = {
        "parsed_text": f"<p>Hello, @{user.username} and @{other_user.username}!</p>",
        "mentions": [],
    }
    add_mentions(request_mock, parsing_result)
    assert (
        f'<a href="{user.get_absolute_url()}">@{user.username}</a>'
        in parsing_result["parsed_text"]
    )
    assert (
        f'<a href="{other_user.get_absolute_url()}">@{other_user.username}</a>'
        in parsing_result["parsed_text"]
    )


def test_util_adds_multiple_mentions_to_parsig_result(request_mock, user, other_user):
    parsing_result = {
        "parsed_text": f"<p>Hello, @{user.username} and @{other_user.username}!</p>",
        "mentions": [],
    }
    add_mentions(request_mock, parsing_result)
    assert parsing_result["mentions"] == [user, other_user]


def test_util_handles_repeated_mentions_of_same_user(request_mock, user):
    parsing_result = {
        "parsed_text": f"<p>Hello, @{user.username} and @{user.username}!</p>",
        "mentions": [],
    }
    add_mentions(request_mock, parsing_result)
    assert parsing_result["mentions"] == [user]


def test_util_skips_mentions_in_links(request_mock, user, snapshot):
    parsing_result = {
        "parsed_text": f'<p>Hello, <a href="/">@{user.username}</a></p>',
        "mentions": [],
    }
    add_mentions(request_mock, parsing_result)
    assert parsing_result["parsed_text"] == (
        f'<p>Hello, <a href="/">@{user.username}</a></p>'
    )
    assert parsing_result["mentions"] == []


def test_util_handles_text_without_mentions(request_mock):
    parsing_result = {"parsed_text": f"<p>Hello, world!</p>", "mentions": []}
    add_mentions(request_mock, parsing_result)
    assert parsing_result["parsed_text"] == ("<p>Hello, world!</p>")
    assert parsing_result["mentions"] == []
